@@ -1,6 +1,5 @@
 from base64 import b64decode
 import copy
-from dataclasses import dataclass
 from math import ceil
 from typing import Any, cast, Optional
 
@@ -31,7 +30,7 @@ from beaker.decorators import (
 )
 from beaker.client.state_decode import decode_state
 from beaker.client.logic_error import LogicException
-from beaker.precompile import AppPrecompile
+from beaker.precompile import AppPrecompile, ProgramAssertion
 
 
 class ApplicationClient:
@@ -56,11 +55,11 @@ class ApplicationClient:
 
         self.approval_binary: Optional[bytes] = None
         self.approval_src_map: Optional[SourceMap] = None
-        self.approval_asserts: dict[int, ProgramAssertion] = {}
+        self.approval_asserts: Optional[dict[int, ProgramAssertion]] = None
 
         self.clear_binary: Optional[bytes] = None
         self.clear_src_map: Optional[SourceMap] = None
-        self.clear_asserts: dict[int, ProgramAssertion] = {}
+        self.clear_asserts: Optional[dict[int, ProgramAssertion]] = None
 
         self.suggested_params = suggested_params
 
@@ -87,17 +86,11 @@ class ApplicationClient:
 
         self.approval_binary = compiled_app.approval._binary
         self.approval_src_map = compiled_app.approval._map
+        self.approval_asserts = compiled_app.approval._asserts
 
         self.clear_binary = compiled_app.clear._binary
         self.clear_src_map = compiled_app.clear._map
-
-        if self.app.approval_program is not None and self.app.clear_program is not None:
-            self.approval_asserts = _gather_asserts(
-                self.app.approval_program, self.approval_src_map
-            )
-            self.clear_asserts = _gather_asserts(
-                self.app.clear_program, self.clear_src_map
-            )
+        self.clear_asserts = compiled_app.clear._asserts
 
     def create(
         self,
@@ -791,31 +784,3 @@ class ApplicationClient:
                 return signer.lsig.address()
 
         raise Exception("No sender provided")
-
-
-@dataclass
-class ProgramAssertion:
-    line: int
-    message: str
-
-
-def _gather_asserts(program: str, src_map: SourceMap) -> dict[int, ProgramAssertion]:
-    asserts: dict[int, ProgramAssertion] = {}
-
-    program_lines = program.split("\n")
-    for idx, line in enumerate(program_lines):
-        # Take only the first bit
-        line = line.split(" ").pop()
-        if line != "assert":
-            continue
-
-        pc = src_map.get_pcs_for_line(idx)[0]
-
-        # TODO: this will be wrong for multiline comments
-        line_before = program_lines[idx - 1]
-        if not line_before.startswith("//"):
-            continue
-
-        asserts[pc] = ProgramAssertion(idx, line_before.strip("// "))
-
-    return asserts
