@@ -2,10 +2,12 @@ import pytest
 import pyteal as pt
 from beaker.state import (
     ApplicationState,
-    DynamicAccountStateValue,
-    DynamicApplicationStateValue,
+    AccountState,
+    ReservedAccountStateValue,
+    ReservedApplicationStateValue,
     ApplicationStateValue,
     AccountStateValue,
+    get_default_for_type,
 )
 
 options = pt.CompileOptions(mode=pt.Mode.Application, version=6)
@@ -20,11 +22,13 @@ ACCOUNT_VAL_TESTS = [
     (pt.Bytes("k"), pt.TealType.bytes, pt.Bytes("abc"), pt.Bytes("def"), None),
     (pt.Bytes("k"), pt.TealType.bytes, pt.Bytes("abc"), pt.Bytes("def"), None),
     (pt.Bytes("k"), pt.TealType.bytes, None, pt.Int(1), pt.TealTypeError),
+    (pt.Bytes("k"), pt.TealType.bytes, pt.Int(1), pt.Bytes("abc"), pt.TealTypeError),
     (pt.Int(123), pt.TealType.bytes, None, pt.Bytes("abc"), pt.TealTypeError),
+    (None, pt.TealType.bytes, None, pt.Bytes("abc"), pt.TealInputError),
 ]
 
 
-DYNAMIC_ACCOUNT_VALUE_TESTS = [
+RESERVED_ACCOUNT_VALUE_TESTS = [
     # stack type, max keys, key gen, key_seed, val, expected error
     (pt.TealType.uint64, 1, None, pt.Bytes("abc"), pt.Int(1), None),
     (pt.TealType.bytes, 1, None, pt.Bytes("abc"), pt.Bytes("abc"), None),
@@ -54,16 +58,6 @@ DYNAMIC_ACCOUNT_VALUE_TESTS = [
 ]
 
 
-def get_default_for_type(stack_type, default):
-    expected_default = default
-    if expected_default is None:
-        if stack_type == pt.TealType.bytes:
-            expected_default = pt.Bytes("")
-        else:
-            expected_default = pt.Int(0)
-    return expected_default
-
-
 @pytest.mark.parametrize("key, stack_type, default, val, error", ACCOUNT_VAL_TESTS)
 def test_local_value(key, stack_type, default, val, error):
     if error is not None:
@@ -76,30 +70,35 @@ def test_local_value(key, stack_type, default, val, error):
 def do_lv_test(key, stack_type, default, val):
     lv = AccountStateValue(stack_type=stack_type, key=key, default=default)
 
-    assert lv.__str__() == f"AccountStateValue {key}"
+    assert lv.__str__() == f"AccountStateValue (Txn Sender) {key}"
 
-    actual = lv.set(val, pt.Txn.sender()).__teal__(options)
+    actual = lv.set(val).__teal__(options)
     expected = pt.App.localPut(pt.Txn.sender(), key, val).__teal__(options)
     with pt.TealComponent.Context.ignoreExprEquality():
         assert actual == expected
 
-    actual = lv.set_default(pt.Txn.sender()).__teal__(options)
+    actual = lv.set_default().__teal__(options)
     expected_default = get_default_for_type(stack_type, default)
     expected = pt.App.localPut(pt.Txn.sender(), key, expected_default).__teal__(options)
     with pt.TealComponent.Context.ignoreExprEquality():
         assert actual == expected
 
-    actual = lv.get(pt.Txn.sender()).__teal__(options)
+    actual = lv.get().__teal__(options)
     expected = pt.App.localGet(pt.Txn.sender(), key).__teal__(options)
     with pt.TealComponent.Context.ignoreExprEquality():
         assert actual == expected
 
+    actual = lv[pt.Txn.accounts[1]].get().__teal__(options)
+    expected = pt.App.localGet(pt.Txn.accounts[1], key).__teal__(options)
+    with pt.TealComponent.Context.ignoreExprEquality():
+        assert actual == expected
+
     actual = lv.__teal__(options)
     expected = pt.App.localGet(pt.Txn.sender(), key).__teal__(options)
     with pt.TealComponent.Context.ignoreExprEquality():
         assert actual == expected
 
-    actual = lv.get_maybe(pt.Txn.sender()).__teal__(options)
+    actual = lv.get_maybe().__teal__(options)
     expected = pt.App.localGetEx(pt.Txn.sender(), pt.Int(0), key).__teal__(options)
     with pt.TealComponent.Context.ignoreExprEquality(), pt.TealComponent.Context.ignoreScratchSlotEquality():
         assert actual == expected
@@ -123,25 +122,37 @@ def do_lv_test(key, stack_type, default, val):
     with pt.TealComponent.Context.ignoreExprEquality(), pt.TealComponent.Context.ignoreScratchSlotEquality():
         assert actual == expected
 
-    actual = lv.delete(pt.Txn.sender()).__teal__(options)
+    actual = lv.exists().__teal__(options)
+    expected = pt.Seq(
+        (v := pt.App.localGetEx(pt.Txn.sender(), pt.Int(0), key)), v.hasValue()
+    ).__teal__(options)
+    with pt.TealComponent.Context.ignoreExprEquality(), pt.TealComponent.Context.ignoreScratchSlotEquality():
+        assert actual == expected
+
+    actual = lv.delete().__teal__(options)
+    expected = pt.App.localDel(pt.Txn.sender(), key).__teal__(options)
+    with pt.TealComponent.Context.ignoreExprEquality():
+        assert actual == expected
+
+    actual = lv.delete().__teal__(options)
     expected = pt.App.localDel(pt.Txn.sender(), key).__teal__(options)
     with pt.TealComponent.Context.ignoreExprEquality():
         assert actual == expected
 
 
 @pytest.mark.parametrize(
-    "stack_type, max_keys, key_gen, key_seed, val, error", DYNAMIC_ACCOUNT_VALUE_TESTS
+    "stack_type, max_keys, key_gen, key_seed, val, error", RESERVED_ACCOUNT_VALUE_TESTS
 )
-def test_dynamic_local_value(stack_type, max_keys, key_gen, key_seed, val, error):
+def test_reserved_local_value(stack_type, max_keys, key_gen, key_seed, val, error):
     if error is not None:
         with pytest.raises(error):
-            do_dynamic_lv_test(stack_type, max_keys, key_gen, key_seed, val)
+            do_reserved_lv_test(stack_type, max_keys, key_gen, key_seed, val)
     else:
-        do_dynamic_lv_test(stack_type, max_keys, key_gen, key_seed, val)
+        do_reserved_lv_test(stack_type, max_keys, key_gen, key_seed, val)
 
 
-def do_dynamic_lv_test(stack_type, max_keys, key_gen, key_seed, val):
-    dlv = DynamicAccountStateValue(
+def do_reserved_lv_test(stack_type, max_keys, key_gen, key_seed, val):
+    dlv = ReservedAccountStateValue(
         stack_type=stack_type, max_keys=max_keys, key_gen=key_gen
     )
 
@@ -153,18 +164,18 @@ def do_dynamic_lv_test(stack_type, max_keys, key_gen, key_seed, val):
     if key_gen is not None:
         key = key_gen(key_seed)
 
-    actual = lv.set(val, pt.Txn.sender()).__teal__(options)
+    actual = lv.set(val).__teal__(options)
     expected = pt.App.localPut(pt.Txn.sender(), key, val).__teal__(options)
     with pt.TealComponent.Context.ignoreExprEquality():
         assert actual == expected
 
-    actual = lv.set_default(pt.Txn.sender()).__teal__(options)
+    actual = lv.set_default().__teal__(options)
     expected_default = get_default_for_type(stack_type, None)
     expected = pt.App.localPut(pt.Txn.sender(), key, expected_default).__teal__(options)
     with pt.TealComponent.Context.ignoreExprEquality():
         assert actual == expected
 
-    actual = lv.get(pt.Txn.sender()).__teal__(options)
+    actual = lv.get().__teal__(options)
     expected = pt.App.localGet(pt.Txn.sender(), key).__teal__(options)
     with pt.TealComponent.Context.ignoreExprEquality():
         assert actual == expected
@@ -174,7 +185,7 @@ def do_dynamic_lv_test(stack_type, max_keys, key_gen, key_seed, val):
     with pt.TealComponent.Context.ignoreExprEquality():
         assert actual == expected
 
-    actual = lv.get_maybe(pt.Txn.sender()).__teal__(options)
+    actual = lv.get_maybe().__teal__(options)
     expected = pt.App.localGetEx(pt.Txn.sender(), pt.Int(0), key).__teal__(options)
     with pt.TealComponent.Context.ignoreExprEquality(), pt.TealComponent.Context.ignoreScratchSlotEquality():
         assert actual == expected
@@ -198,7 +209,14 @@ def do_dynamic_lv_test(stack_type, max_keys, key_gen, key_seed, val):
     with pt.TealComponent.Context.ignoreExprEquality(), pt.TealComponent.Context.ignoreScratchSlotEquality():
         assert actual == expected
 
-    actual = lv.delete(pt.Txn.sender()).__teal__(options)
+    actual = lv.exists().__teal__(options)
+    expected = pt.Seq(
+        (v := pt.App.localGetEx(pt.Txn.sender(), pt.Int(0), key)), v.hasValue()
+    ).__teal__(options)
+    with pt.TealComponent.Context.ignoreExprEquality(), pt.TealComponent.Context.ignoreScratchSlotEquality():
+        assert actual == expected
+
+    actual = lv.delete().__teal__(options)
     expected = pt.App.localDel(pt.Txn.sender(), key).__teal__(options)
     with pt.TealComponent.Context.ignoreExprEquality():
         assert actual == expected
@@ -216,6 +234,7 @@ APPLICATION_VAL_TESTS = [
     (pt.Bytes("k"), pt.TealType.bytes, None, pt.Int(1), False, pt.TealTypeError),
     (pt.Int(123), pt.TealType.bytes, None, pt.Bytes("abc"), False, pt.TealTypeError),
     (pt.Bytes("k"), pt.TealType.bytes, None, pt.Bytes("abc"), True, pt.TealInputError),
+    (None, pt.TealType.bytes, None, pt.Bytes("abc"), False, pt.TealInputError),
 ]
 
 
@@ -252,8 +271,17 @@ def do_gv_test(key, stack_type, default, val, static):
 
     actual = lv.set_default().__teal__(options)
     expected_default = get_default_for_type(stack_type, default)
-    expected = pt.App.globalPut(key, expected_default).__teal__(options)
-    with pt.TealComponent.Context.ignoreExprEquality():
+
+    if static:
+        expected = pt.Seq(
+            v := pt.App.globalGetEx(pt.Int(0), key),
+            pt.Assert(pt.Not(v.hasValue())),
+            pt.App.globalPut(key, expected_default),
+        ).__teal__(options)
+    else:
+        expected = pt.App.globalPut(key, expected_default).__teal__(options)
+
+    with pt.TealComponent.Context.ignoreExprEquality(), pt.TealComponent.Context.ignoreScratchSlotEquality():
         assert actual == expected
 
     actual = lv.get().__teal__(options)
@@ -308,13 +336,20 @@ def do_gv_test(key, stack_type, default, val, static):
     with pt.TealComponent.Context.ignoreExprEquality(), pt.TealComponent.Context.ignoreScratchSlotEquality():
         assert actual == expected
 
+    actual = lv.exists().__teal__(options)
+    expected = pt.Seq((v := pt.App.globalGetEx(pt.Int(0), key)), v.hasValue()).__teal__(
+        options
+    )
+    with pt.TealComponent.Context.ignoreExprEquality(), pt.TealComponent.Context.ignoreScratchSlotEquality():
+        assert actual == expected
+
     actual = lv.delete().__teal__(options)
     expected = pt.App.globalDel(key).__teal__(options)
     with pt.TealComponent.Context.ignoreExprEquality():
         assert actual == expected
 
 
-DYNAMIC_APPLICATION_VALUE_TESTS = [
+RESERVED_APPLICATION_VALUE_TESTS = [
     # stack type, max keys, key gen, key_seed, val, expected error
     (pt.TealType.uint64, 1, None, pt.Bytes("abc"), pt.Int(1), None),
     (pt.TealType.bytes, 1, None, pt.Bytes("abc"), pt.Bytes("abc"), None),
@@ -346,18 +381,18 @@ DYNAMIC_APPLICATION_VALUE_TESTS = [
 
 @pytest.mark.parametrize(
     "stack_type, max_keys, key_gen, key_seed, val, error",
-    DYNAMIC_APPLICATION_VALUE_TESTS,
+    RESERVED_APPLICATION_VALUE_TESTS,
 )
-def test_dynamic_global_value(stack_type, max_keys, key_gen, key_seed, val, error):
+def test_reserved_global_value(stack_type, max_keys, key_gen, key_seed, val, error):
     if error is not None:
         with pytest.raises(error):
-            do_dynamic_gv_test(stack_type, max_keys, key_gen, key_seed, val)
+            do_reserved_gv_test(stack_type, max_keys, key_gen, key_seed, val)
     else:
-        do_dynamic_gv_test(stack_type, max_keys, key_gen, key_seed, val)
+        do_reserved_gv_test(stack_type, max_keys, key_gen, key_seed, val)
 
 
-def do_dynamic_gv_test(stack_type, max_keys, key_gen, key_seed, val):
-    dlv = DynamicApplicationStateValue(
+def do_reserved_gv_test(stack_type, max_keys, key_gen, key_seed, val):
+    dlv = ReservedApplicationStateValue(
         stack_type=stack_type, max_keys=max_keys, key_gen=key_gen
     )
 
@@ -427,6 +462,13 @@ def do_dynamic_gv_test(stack_type, max_keys, key_gen, key_seed, val):
     with pt.TealComponent.Context.ignoreExprEquality(), pt.TealComponent.Context.ignoreScratchSlotEquality():
         assert actual == expected
 
+    actual = lv.exists().__teal__(options)
+    expected = pt.Seq((v := pt.App.globalGetEx(pt.Int(0), key)), v.hasValue()).__teal__(
+        options
+    )
+    with pt.TealComponent.Context.ignoreExprEquality(), pt.TealComponent.Context.ignoreScratchSlotEquality():
+        assert actual == expected
+
     actual = lv.delete().__teal__(options)
     expected = pt.App.globalDel(key).__teal__(options)
     with pt.TealComponent.Context.ignoreExprEquality():
@@ -453,6 +495,21 @@ def test_application_state():
     assert astate.num_byte_slices == 1
     assert astate.num_uints == 1
 
+    statevals["c"] = ReservedApplicationStateValue(pt.TealType.uint64, max_keys=64)
+    with pytest.raises(Exception):
+        ApplicationState(statevals)
+
 
 def test_account_state():
-    pass
+    statevals = {
+        "a": AccountStateValue(pt.TealType.uint64),
+        "b": AccountStateValue(pt.TealType.bytes),
+    }
+    astate = AccountState(statevals)
+
+    assert astate.num_byte_slices == 1
+    assert astate.num_uints == 1
+
+    statevals["c"] = ReservedAccountStateValue(pt.TealType.uint64, max_keys=16)
+    with pytest.raises(Exception):
+        AccountState(statevals)
